"""An extended version of SAM learner that is able to handle predicates inconsistency."""
from collections import defaultdict
from typing import Optional, Dict, Set, NoReturn, Iterable, List

from pddl.pddl import Action, Domain, Effect

from .core import ExtendedMatcher, extract_effects, LightProxyActionGenerator
from .sam_learner import SAMLearner
from .sam_models import Mode, GroundedAction, State, ComparablePredicate, Trajectory, SignatureType, TrajectoryComponent


def sort_predicates(predicates: Iterable[ComparablePredicate]) -> List[ComparablePredicate]:
    """Sorts the predicate list so that it could be compared to other lists.

    :param predicates: the predicates to sort.
    :return: the sorted predicate list.
    """
    return sorted(
        predicates, key=lambda predicate: (predicate.name, str(predicate.signature)))


class ESAMLearner(SAMLearner):
    """Extension to the SAM Learning algorithm."""

    matcher: ExtendedMatcher
    proxy_action_generator: LightProxyActionGenerator
    # Maps an action name to a dictionary of effect predicate names to their possible bindings.
    add_effect_cnfs: Dict[str, Dict[str, Set[ComparablePredicate]]]
    delete_effect_cnfs: Dict[str, Dict[str, Set[ComparablePredicate]]]
    # Maps an action name to a dictionary of preconditions predicate names to their possible bindings.
    actions_possible_preconditions: Dict[str, Set[ComparablePredicate]]
    learned_actions_signatures: Dict[str, SignatureType]

    def __init__(
            self, working_directory_path: Optional[str] = None, domain_file_name: str = "domain.pddl",
            mode: Mode = "production", domain: Optional[Domain] = None, known_actions: Dict[str, Action] = {}):
        super().__init__(working_directory_path, domain_file_name, mode, domain, known_actions)
        self.proxy_action_generator = LightProxyActionGenerator()
        self.add_effect_cnfs = {}
        self.delete_effect_cnfs = {}
        self.actions_possible_preconditions = {}
        self.learned_actions_signatures = {}
        if mode == "development":
            self.matcher = ExtendedMatcher(domain=domain)
            return

        domain_path = self.working_directory_path / domain_file_name
        self.matcher = ExtendedMatcher(domain_path=str(domain_path))

    def _add_predicates_cnfs(
            self, predicates: List[ComparablePredicate]) -> Dict[str, Set[ComparablePredicate]]:
        """Adds fluents to a CNF clause when needed.

        :param predicates: the predicates that have been currently observed.
        :return: the dictionary of the fluents after the new information was added.
        """
        predicates_cnf = defaultdict(set)
        for predicate in predicates:
            for index, signature_item in enumerate(predicate.signature):
                if type(signature_item[1]) is tuple:
                    continue

                predicate.signature[index] = (signature_item[0], (signature_item[1],))

            predicates_cnf[predicate.name].add(predicate)

        return predicates_cnf

    def _update_preconditions(self, action_name: str,
                              possible_preconditions: List[ComparablePredicate]) -> NoReturn:
        """Update the precondition for an action that had already been observed.

        :param action_name: the name of the action that is currently being learned.
        :param possible_preconditions: the fluents that were observed in the current trajectory triplet.
        """
        stored_preconditions = self.actions_possible_preconditions[action_name]
        self.logger.debug("Removing predicates that don't exist in the current trajectory triplet.")
        remaining_preconditions = [predicate for predicate in possible_preconditions if
                                   predicate in stored_preconditions]
        self.actions_possible_preconditions[action_name] = set(remaining_preconditions)

    def _update_effects_cnfs(self, new_lifted_effects: List[ComparablePredicate],
                             effects_to_update: Dict[str, Set[ComparablePredicate]]) -> NoReturn:
        """Update the CNF clauses of an already observed action.

        :param action_name: the name of the action that is being updated.
        :param new_lifted_effects: the newly observed lifted effects observed for the action.
        :param effects_to_update: the current CNF clauses that the action has.
        """
        new_effects_cnfs = self._add_predicates_cnfs(new_lifted_effects)
        not_encountered_cnfs = {}
        for predicate_name, new_possible_effect_cnfs in new_effects_cnfs.items():
            if predicate_name not in effects_to_update:
                self.logger.debug("Adding a new effect that hadn't been encountered before.")
                not_encountered_cnfs[predicate_name] = new_possible_effect_cnfs

            else:
                self.logger.debug("Removing redundant CNF clauses from the effect clauses of the predicate.")
                previous_effects = effects_to_update[predicate_name]
                effects_to_update[predicate_name] = new_possible_effect_cnfs.intersection(previous_effects)

        effects_to_update.update(not_encountered_cnfs)

    def add_new_action(
            self, grounded_action: GroundedAction, previous_state: State, next_state: State) -> NoReturn:
        """Learns the model of an action that was observed for the first time.

        :param grounded_action: the grounded action that was observed in the trajectory triplet.
        :param previous_state: the state that was observed prior to the action's exection.
        :param next_state: the state that was observed after the action's execution.
        """
        action_name = grounded_action.lifted_action_name
        self.logger.info(f"Action {action_name} encountered for the first time! Adding its data to the data structure.")
        action_signature = grounded_action.lifted_signature
        self.learned_actions_signatures[action_name] = action_signature
        possible_preconditions = self.matcher.get_possible_literal_matches(grounded_action,
                                                                           previous_state.facts)
        self.actions_possible_preconditions[action_name] = set(possible_preconditions)

        grounded_add_effects, grounded_del_effects = extract_effects(previous_state, next_state)
        lifted_add_effects = self.matcher.get_possible_literal_matches(grounded_action, grounded_add_effects)
        self.add_effect_cnfs[action_name] = self._add_predicates_cnfs(lifted_add_effects)
        lifted_del_effects = self.matcher.get_possible_literal_matches(grounded_action, grounded_del_effects)
        self.delete_effect_cnfs[action_name] = self._add_predicates_cnfs(lifted_del_effects)
        self.logger.debug(f"Finished adding {action_name} information to the data structure")

    def update_action(self, grounded_action: GroundedAction, previous_state: State, next_state: State) -> NoReturn:
        """Updates an action that was observed at least once already.

        :param grounded_action: the grounded action that was executed according to the trajectory.
        :param previous_state: the state that the action was executed on.
        :param next_state: the state that was created after executing the action on the previous
            state.
        """
        action_name = grounded_action.lifted_action_name
        self.logger.info(f"Starting to update the action - {action_name}")
        observed_pre_state_predicates = self.matcher.get_possible_literal_matches(grounded_action, previous_state.facts)
        self._update_preconditions(action_name, observed_pre_state_predicates)

        grounded_add_effects, grounded_del_effects = extract_effects(previous_state, next_state)
        new_lifted_add_effects = self.matcher.get_possible_literal_matches(grounded_action, grounded_add_effects)
        new_lifted_delete_effects = self.matcher.get_possible_literal_matches(grounded_action, grounded_del_effects)
        self._update_effects_cnfs(new_lifted_add_effects, self.add_effect_cnfs[action_name])
        self._update_effects_cnfs(new_lifted_delete_effects, self.delete_effect_cnfs[action_name])
        self.logger.debug(f"Done updating the action - {action_name}")

    def _is_proxy_action(self, action_name: str) -> bool:
        """Validate whether or not an action is supposed to be a proxy action due to the fact that it has ambiguous
            effects.

        :param action_name: the name of the action that is currently being tested.
        :return: whether or not an action is a proxy action.
        """
        action_add_effects = self.add_effect_cnfs[action_name]
        action_del_effects = self.delete_effect_cnfs[action_name]
        return any([len(cnf) > 1 for cnf in action_add_effects.values()]) or \
               any([len(cnf) > 1 for cnf in action_del_effects.values()])

    def create_proxy_actions(self) -> Dict[str, Action]:
        """Create the proxy actions for the cases where there is ambiguity in the learning process.

        :return: the actions that the model learned through its execution stage.
        """
        learned_actions = {}
        for action_name in self.actions_possible_preconditions:
            if not self._is_proxy_action(action_name):
                self.logger.debug(f"Action - {action_name} has no ambiguities, creating regular action.")
                action = self._create_action_from_cnf(action_name)
                learned_actions[action_name] = action

            else:
                # In the light version of the proxy action generator we don't have to remove the constants.
                self.logger.debug(f"Creating proxy actions for the action - {action_name}")
                proxy_actions = self.proxy_action_generator.create_proxy_actions(
                    action_name=action_name,
                    action_signature=self.learned_actions_signatures[action_name],
                    surely_preconditions=self.actions_possible_preconditions[action_name],
                    add_effect_cnfs=self.add_effect_cnfs[action_name],
                    delete_effect_cnfs=self.delete_effect_cnfs[action_name]
                )
                for action in proxy_actions:
                    learned_actions[action.name] = action

        return learned_actions

    def _create_action_from_cnf(self, action_name: str) -> Action:
        """Create the action object from the CNF clauses collected through the algorithm's execution.

        :param action_name: the name of the action that is currently being created.
        :return: the action that was created from the CNF clauses (not proxy action).
        """
        action_add_effect_cnf = self.add_effect_cnfs[action_name]
        action_delete_effect_cnf = self.delete_effect_cnfs[action_name]
        add_effects = set()
        delete_effects = set()

        for add_fluents in action_add_effect_cnf.values():
            add_effects.update(add_fluents)

        for delete_fluents in action_delete_effect_cnf.values():
            delete_effects.update(delete_fluents)

        effect = Effect()
        effect.addlist = add_effects
        effect.dellist = delete_effects

        return Action(name=action_name, signature=self.learned_actions_signatures[action_name],
                      precondition=list(self.actions_possible_preconditions[action_name]), effect=effect)

    def handle_single_trajectory_component(self, component: TrajectoryComponent) -> NoReturn:
        """Handles a single trajectory component as a part of the learning process.

        :param component: the trajectory component that is being handled at the moment.
        """
        previous_state = component.previous_state
        grounded_action = component.grounded_action
        next_state = component.next_state
        action_name = grounded_action.lifted_action_name
        if self._is_known_action(action_name):
            self.logger.debug(f"The action - {action_name} is already known to the agent.")
            return

        self._verify_parameter_duplication(grounded_action)
        self.action_to_triplets_histogram[action_name] += 1
        if action_name not in self.learned_actions_signatures:
            self.add_new_action(grounded_action, previous_state, next_state)

        else:
            self.update_action(grounded_action, previous_state, next_state)

    def learn_action_model(self, trajectories: List[Trajectory]) -> Domain:
        """Learn the SAFE action model from the input trajectories.

        :param trajectories: the list of trajectories that are used to learn the safe action model.
        :return: a domain containing the actions that were learned.
        """
        self.logger.info("Starting to learn the action model!")
        for trajectory in trajectories:
            for component in trajectory:
                self.handle_single_trajectory_component(component)

        learned_actions = self.create_proxy_actions()
        learned_actions.update(self.known_actions)
        self.learned_domain.actions = learned_actions
        return self.learned_domain



"""Calculates the number of consistent models are extractable from the learned domain."""
import csv
import logging
from pathlib import Path
from typing import List, Set, NoReturn

from pddl.parser import Parser
from pddl.pddl import Domain, Action

from sam_learner.sam_models import ComparablePredicate

STATISTICS_COLUMNS_NAMES = ["domain_name", "domain_path", "action_name", "number_consistent_models"]


def calculate_number_consistent_models(
		maybe_preconditions: List[ComparablePredicate],
		maybe_effects: Set[ComparablePredicate] = frozenset()) -> int:
	"""calculates the number of consistent models that are available based on the learned preconditions and effects.

	:param maybe_preconditions: the possible preconditions for the learned action.
	:param maybe_effects: the possible effects for the learned action.
	:return: the total number of possible action models that could be learned for the specific action.
	"""
	return pow(2, len(maybe_preconditions)) * pow(2, len(maybe_effects))


class SafeConsistentModelsCalculator:

	def __init__(self):
		self.logger = logging.getLogger(__name__)

	def calculate_safe_models_consistency(
			self, generated_domains_directory_path: Path, output_statistics_path: str,
			known_actions: List[str]) -> NoReturn:
		"""

		:param generated_domains_directory_path:
		:param output_statistics_path:
		:param known_actions:
		:return:
		"""
		self.logger.info("Logging the number of consistent models for the safe action models that was learned.")
		consistency_statistics = []
		for generated_domain_path in generated_domains_directory_path.glob("*.pddl"):
			self.logger.debug(f"Logging the consistency statistics for the domain - {generated_domain_path.stem}")
			domain: Domain = Parser(generated_domain_path).parse_domain()
			learned_actions = [action_name for action_name in domain.actions if action_name not in known_actions]
			for action_name in learned_actions:
				action_data: Action = domain.actions[action_name]
				action_preconditions = action_data.precondition
				action_consistency = {
					"domain_name": domain.name,
					"domain_path": generated_domain_path,
					"action_name": action_name,
					"number_consistent_models": calculate_number_consistent_models(action_preconditions)
				}
				consistency_statistics.append(action_consistency)

		with open(output_statistics_path, "wt") as csv_file:
			writer = csv.DictWriter(csv_file, fieldnames=STATISTICS_COLUMNS_NAMES)
			writer.writeheader()
			for data in consistency_statistics:
				writer.writerow(data)



"""Module that contains the SAM learner code for single agent problems."""
import logging
from collections import defaultdict
from pathlib import Path
from typing import List, Optional, NoReturn, Dict, Set

from pddl.pddl import Domain, Action, Effect
from pyperplan import Parser

from sam_learner.core import PredicatesMatcher, extract_effects, extract_maybe_fluents
from sam_learner.sam_models import GroundedAction, State, TrajectoryComponent, Trajectory, Mode, ComparablePredicate


def contains_duplicates(tested_list: List[str]) -> bool:
    """checks if the list contains duplicates.

    :param tested_list: the list to test for duplicates.
    :return: whether or not the list contains duplicates.
    """
    return len(set(tested_list)) != len(tested_list)


class SAMLearner:
    """Class that represents the safe action model learner algorithm."""

    logger: logging.Logger
    working_directory_path: Path
    trajectories: List[Trajectory]
    learned_domain: Domain
    matcher: PredicatesMatcher
    known_actions: Dict[str, Action]
    maybe_effects: Dict[str, Set[ComparablePredicate]]
    action_to_triplets_histogram: Dict[str, int]
    action_with_duplicate_param_calls: Dict[str, int]

    def __init__(
            self, working_directory_path: Optional[str] = None, domain_file_name: str = "domain.pddl",
            mode: Mode = "production",
            domain: Optional[Domain] = None, known_actions: Dict[str, Action] = {}):
        self.logger = logging.getLogger(__name__)
        self.known_actions = known_actions
        self.maybe_effects = defaultdict(set)
        self.action_to_triplets_histogram = defaultdict(int)
        self.action_with_duplicate_param_calls = defaultdict(int)
        if mode == "development":
            self.matcher = PredicatesMatcher(domain=domain)
            self.learned_domain = domain
            return

        self.working_directory_path = Path(working_directory_path)
        domain_path = self.working_directory_path / domain_file_name
        self.learned_domain = Parser(domain_path).parse_domain(read_from_file=True)
        self.learned_domain.actions = {}
        self.matcher = PredicatesMatcher(domain_path=str(domain_path))
        if known_actions is not None:
            self.learned_domain.actions = {
                name: action for name, action in known_actions.items()
            }

    def handle_action_effects(
            self, grounded_action: GroundedAction, previous_state: State, next_state: State) -> Effect:
        """Finds the effects generated from the previous and the next state on this current step.

        :param grounded_action: the grounded action that was executed according to the trajectory.
        :param previous_state: the state that the action was executed on.
        :param next_state: the state that was created after executing the action on the previous
            state.
        :return: the effect containing the add and del list of predicates.
        """
        grounded_add_effects, grounded_del_effects = extract_effects(previous_state, next_state)
        action_effect = Effect()
        action_effect.addlist = action_effect.addlist.union(self.matcher.get_possible_literal_matches(
            grounded_action, grounded_add_effects))
        action_effect.dellist = action_effect.dellist.union(self.matcher.get_possible_literal_matches(
            grounded_action, grounded_del_effects))

        self.handle_maybe_effects(grounded_action, previous_state, next_state)

        return action_effect

    def handle_maybe_effects(
            self, grounded_action: GroundedAction, previous_state: State, next_state: State) -> NoReturn:
        """Extracts the maybe effects that are caused by the intersection between the previous and the next state.

        :param grounded_action: the currently used grounded action.
        :param previous_state: the state that the action was executed on.
        :param next_state: the state that was created after executing the action on the previous
            state.
        """
        maybe_effect_fluents = extract_maybe_fluents(previous_state, next_state)
        maybe_effects = self.matcher.get_possible_literal_matches(grounded_action, maybe_effect_fluents)
        action_name = grounded_action.lifted_action_name
        if action_name in self.maybe_effects:
            self.maybe_effects[action_name].intersection_update(maybe_effects)
        else:
            self.maybe_effects[action_name].update(maybe_effects)

    def add_new_action(
            self, grounded_action: GroundedAction, previous_state: State, next_state: State) -> NoReturn:
        """Create a new action in the domain.

        :param grounded_action: the grounded action that was executed according to the trajectory.
        :param previous_state: the state that the action was executed on.
        :param next_state: the state that was created after executing the action on the previous
            state.
        """
        self.logger.info(f"Adding the action {grounded_action.activated_action_representation} "
                         f"to the domain.")
        new_action = Action(name=grounded_action.lifted_action_name,
                            signature=grounded_action.lifted_signature,
                            precondition=[],
                            effect=None)

        # adding the preconditions each predicate is grounded in this stage.
        possible_preconditions = self.matcher.get_possible_literal_matches(grounded_action,
                                                                           previous_state.facts)
        new_action.precondition = list(set(possible_preconditions))

        action_effect = self.handle_action_effects(grounded_action, previous_state, next_state)
        new_action.effect = action_effect
        self.learned_domain.actions[new_action.name] = new_action
        self.logger.debug(
            f"Finished adding the action {grounded_action.activated_action_representation}.")

    def _is_known_action(self, action_name: str) -> bool:
        """Check whether or not the input action is an action that the agent shouldn't learn.

        :param action_name: the name of the action that is currently observed in the trajectory.
        :return: whether or not the action is already known to the agent.
        """
        self.logger.info(f"Updating the action - {action_name}")
        if action_name in self.known_actions:
            self.logger.debug(f"The action {action_name} is already known to the agent. Skipping!")
            return True

        return False

    def update_action(
            self, grounded_action: GroundedAction, previous_state: State, next_state: State) -> NoReturn:
        """Create a new action in the domain.

        :param grounded_action: the grounded action that was executed according to the trajectory.
        :param previous_state: the state that the action was executed on.
        :param next_state: the state that was created after executing the action on the previous
            state.
        """
        action_name = grounded_action.lifted_action_name
        if self._is_known_action(action_name):
            return

        current_action: Action = self.learned_domain.actions[action_name]
        self._update_action_preconditions(current_action, grounded_action, previous_state)

        action_effect: Effect = self.handle_action_effects(
            grounded_action, previous_state, next_state)
        current_action.effect.addlist = current_action.effect.addlist.union(action_effect.addlist)
        current_action.effect.dellist = current_action.effect.dellist.union(action_effect.dellist)
        self.logger.debug(f"Done updating the action - {grounded_action.lifted_action_name}")

    def _update_action_preconditions(
            self, current_action: Action, grounded_action: GroundedAction, previous_state: State) -> NoReturn:
        """Updates the preconditions of an action after it was observed at least once.

        :param current_action: the action that is being observed.
        :param grounded_action: the grounded action that is being executed in the trajectory component.
        :param previous_state: the state that was seen prior to the action's execution.
        """
        model_preconditions = current_action.precondition.copy()
        possible_preconditions = self.matcher.get_possible_literal_matches(
            grounded_action, previous_state.facts)
        if len(possible_preconditions) > 0:
            for precondition in model_preconditions:
                if precondition not in possible_preconditions:
                    current_action.precondition.remove(precondition)
        else:
            self.logger.warning(f"while handling the action {grounded_action.activated_action_representation} "
                                f"inconsistency occurred, since we do not allow for duplicates we do not update the "
                                f"preconditions.")

    def _verify_parameter_duplication(self, grounded_action: GroundedAction) -> NoReturn:
        """Verifies if the action was called with duplicated objects in a trajectory component.

        :param grounded_action: the grounded action observed in the trajectory triplet.
        """
        predicate_objects = [signature_item[0] for signature_item in grounded_action.grounded_signature]
        if contains_duplicates(predicate_objects):
            self.action_with_duplicate_param_calls[grounded_action.lifted_action_name] += 1

    def handle_single_trajectory_component(self, component: TrajectoryComponent) -> NoReturn:
        """Handles a single trajectory component as a part of the learning process.

        :param component: the trajectory component that is being handled at the moment.
        """
        previous_state = component.previous_state
        grounded_action = component.grounded_action
        next_state = component.next_state
        action_name = grounded_action.lifted_action_name
        if self._is_known_action(action_name):
            self.logger.debug(f"The action - {action_name} is already known to the agent.")
            return

        self._verify_parameter_duplication(grounded_action)
        self.action_to_triplets_histogram[action_name] += 1
        if action_name not in self.learned_domain.actions:
            self.add_new_action(grounded_action, previous_state, next_state)

        else:
            self.update_action(grounded_action, previous_state, next_state)

    def get_actions_appearance_histogram(self) -> Dict[str, int]:
        """Returns the histogram value of the learned actions.

        :return: the histogram of the learned actions.
        """
        return self.action_to_triplets_histogram

    def get_actions_with_duplicated_parameters_histogram(self) -> Dict[str, int]:
        """Returns the histogram value of the learned actions with the duplicated objects.

        :return: the histogram of the learned actions where their usage contained duplicated objects.
        """
        return self.action_with_duplicate_param_calls

    def learn_action_model(self, trajectories: List[Trajectory]) -> Domain:
        """Learn the SAFE action model from the input trajectories.

        :param trajectories: the list of trajectories that are used to learn the safe action model.
        :return: a domain containing the actions that were learned.
        """
        self.logger.info("Starting to learn the action model!")
        for trajectory in trajectories:
            for component in trajectory:
                self.handle_single_trajectory_component(component)

        return self.learned_domain



import logging
import os
import re
import subprocess
import time
from pathlib import Path
from typing import NoReturn

script_headers = """#!/bin/bash

################################################################################################
### sbatch configuration parameters must start with #SBATCH and must precede any other commands.
### To ignore, just add another # - like so: ##SBATCH
################################################################################################

#SBATCH --partition main			### specify partition name where to run a job. short: 7 days limit; gtx1080: 7 days; debug: 2 hours limit and 1 job at a time
#SBATCH --time 0-03:30:00			### limit the time of job running. Make sure it is not greater than the partition time limit!! Format: D-H:MM:SS
#SBATCH --job-name downward_planner_job			### name of the job
#SBATCH --output job-%J.out			### output log for running job - %J for job number
##SBATCH --mail-user=aaa.bbb@ccc	### user's email for sending job status messages
##SBATCH --mail-type=END			### conditions for sending the email. ALL,BEGIN,END,FAIL, REQUEU, NONE

##SBATCH --mem=32G				### amount of RAM memory
##SBATCH --cpus-per-task=6			### number of CPU cores

module load anaconda
conda info --envs
source activate safe_action_learner
"""

script_execution_headers = """

directory_path={test_set_directory_path}
files_regex_path="$directory_path/{problems_regex}"
fast_downward_dir_path="/home/mordocha/downward"
fast_downward_file_path="$fast_downward_dir_path/fast-downward.py"
domain_path={domain_file_path}

"""

script_execution_lines = """

for FILE in $files_regex_path; do
  filename=$(basename -- "$FILE")
  name="${filename%.*}"
  echo "Processing the file - ${filename}:"
  if [[ $name == combined_domain ]]; then
    continue
  fi
  if test -f "${directory_path}/${name}_plan.solution"; then
    continue
  else
    python "$fast_downward_file_path" --overall-time-limit "1m" --sas-file "${directory_path}/${name}-output.sas" --plan-file "${directory_path}/${name}_plan.solution" "$domain_path" "$FILE" --evaluator "hff=ff()" --evaluator "hcea=cea()" --search "lazy_greedy([hff, hcea], preferred=[hff, hcea])"
    sleep 1
    rm "${directory_path}/${name}-output.sas"
  fi
done

"""

BATCH_JOB_SUBMISSION_REGEX = re.compile(b"Submitted batch job (?P<batch_id>\d+)")


class FastDownwardSolver:
	logger: logging.Logger

	def __init__(self):
		self.logger = logging.getLogger(__name__)

	def write_batch_and_execute_solver(
			self, output_file_path: Path, test_set_directory_path: Path, problems_regex: str,
			domain_file_path: str) -> NoReturn:
		"""

		:param output_file_path:
		:param test_set_directory_path:
		:param problems_regex:
		:param domain_file_path:
		:return:
		"""
		os.chdir(output_file_path.parent)
		self.logger.info("Starting to solve the input problems using fast downward solver.")
		completed_file_str = script_headers
		completed_file_str += script_execution_headers.format(
			test_set_directory_path=str(test_set_directory_path),
			problems_regex=problems_regex,
			domain_file_path=domain_file_path
		)
		completed_file_str += script_execution_lines

		with open(output_file_path, "wt") as run_script_file:
			run_script_file.write(completed_file_str)

		submission_str = subprocess.check_output(["sbatch", str(output_file_path)])
		match = BATCH_JOB_SUBMISSION_REGEX.match(submission_str)
		batch_id = match.group("batch_id")
		# waiting for fast downward process to start
		time.sleep(1)
		execution_state = subprocess.check_output(["squeue", "--me"])
		while batch_id in execution_state:
			self.logger.debug(f"Solver with the id - {batch_id} is still running...")
			execution_state = subprocess.check_output(["squeue", "--me"])
			time.sleep(1)
			continue

		self.logger.info("Solver finished its execution!")

		self.logger.debug("Cleaning the sbatch file from the problems directory.")
		os.remove(output_file_path)
		return



"""Module to calculate the precision and recall of our learning algorithm."""
import logging
from typing import List, Any, Dict, NoReturn

from pddl.pddl import Domain

from sam_learner.sam_models import Trajectory, ComparablePredicate


def calculate_true_positive_value(
		learned_predicates: List[ComparablePredicate], expected_predicates: List[ComparablePredicate]) -> int:
	"""

	:param learned_predicates:
	:param expected_predicates:
	:return:
	"""
	return len([predicate for predicate in learned_predicates if predicate in expected_predicates])


def calculate_false_positive_value(
		learned_predicates: List[ComparablePredicate], expected_predicates: List[ComparablePredicate]) -> int:
	"""

	:param learned_predicates:
	:param expected_predicates:
	:return:
	"""
	return len(learned_predicates) - calculate_true_positive_value(learned_predicates, expected_predicates)


def calculate_false_negative_value(
		learned_predicates: List[ComparablePredicate], expected_predicates: List[ComparablePredicate]) -> int:
	"""

	:param learned_predicates:
	:param expected_predicates:
	:return:
	"""
	return len(expected_predicates) - calculate_true_positive_value(learned_predicates, expected_predicates)


def calculate_recall(
		learned_predicates: List[ComparablePredicate], actual_predicates: List[ComparablePredicate]) -> float:
	"""

	:param learned_predicates:
	:param actual_predicates:
	:return:
	"""
	if len(learned_predicates) == 0:
		return 0

	true_positives = calculate_true_positive_value(learned_predicates, actual_predicates)
	false_negatives = calculate_false_negative_value(learned_predicates, actual_predicates)
	return true_positives / (true_positives + false_negatives)


def calculate_precision(
		learned_predicates: List[ComparablePredicate], actual_predicates: List[ComparablePredicate]) -> float:
	"""

	:param learned_predicates:
	:param actual_predicates:
	:return:
	"""
	if len(learned_predicates) == 0:
		return 0

	true_positives = calculate_true_positive_value(learned_predicates, actual_predicates)
	false_positives = calculate_false_positive_value(learned_predicates, actual_predicates)
	return true_positives / (true_positives + false_positives)


class PrecisionRecallCalculator:
	logger: logging.Logger
	expected_domain: Domain

	def __init__(self, expected_domain: Domain):
		self.logger = logging.getLogger(__name__)
		self.expected_domain = expected_domain

	def calculate_precision_recall_for_unknown_actions(
			self, action_name: str, num_trajectories: int, num_triplets: int) -> Dict[str, float]:
		"""

		:param action_name:
		:param num_trajectories:
		:param num_triplets:
		:return:
		"""
		actual_action_preconditions = self.expected_domain.actions[action_name].precondition
		actual_action_add_effects = list(self.expected_domain.actions[action_name].effect.addlist)
		actual_action_delete_effects = list(self.expected_domain.actions[action_name].effect.dellist)
		return {
			"domain_name": self.expected_domain.name,
			"num_trajectories": num_trajectories,
			"num_trajectory_triplets": num_triplets,
			"learned_action_name": action_name,
			"learned_preconditions": None,
			"learned_add_effects": None,
			"learned_delete_effects": None,
			"ground_truth_preconditions": ",".join([str(precondition) for precondition in actual_action_preconditions]),
			"ground_truth_add_effects": ",".join([str(add_effect) for add_effect in actual_action_add_effects]),
			"ground_truth_delete_effects": ",".join([str(del_effect) for del_effect in actual_action_delete_effects]),
			"preconditions_precision": 0,
			"add_effects_precision": 1,
			"delete_effects_precision": 1,
			"preconditions_recall": 1,
			"add_effects_recall": 0,
			"delete_effects_recall": 0,
		}

	def calculate_action_precision_recall(
			self, learned_model: Domain, action_name: str) -> Dict[str, float]:
		"""

		:param learned_model:
		:param action_name:
		:return:
		"""
		learned_action_preconditions = learned_model.actions[action_name].precondition
		learned_action_add_effects = list(learned_model.actions[action_name].effect.addlist)
		learned_action_delete_effects = list(learned_model.actions[action_name].effect.dellist)
		actual_action_preconditions = [ComparablePredicate(predicate=p) for p in self.expected_domain.actions[action_name].precondition]
		actual_action_add_effects = [ComparablePredicate(predicate=p) for p in self.expected_domain.actions[action_name].effect.addlist]
		actual_action_delete_effects = [ComparablePredicate(predicate=p) for p in self.expected_domain.actions[action_name].effect.dellist]
		return {
			"preconditions_precision": calculate_precision(learned_action_preconditions, actual_action_preconditions),
			"add_effects_precision": calculate_precision(learned_action_add_effects, actual_action_add_effects),
			"delete_effects_precision": calculate_precision(learned_action_delete_effects,
															actual_action_delete_effects),
			"preconditions_recall": calculate_recall(learned_action_preconditions, actual_action_preconditions),
			"add_effects_recall": calculate_recall(learned_action_add_effects, actual_action_add_effects),
			"delete_effects_recall": calculate_recall(learned_action_delete_effects, actual_action_delete_effects),
		}

	def extract_learned_action_data_vs_ground_truth(
			self, learned_model: Domain, action_name: str) -> Dict[str, str]:
		"""

		:param learned_model:
		:param action_name:
		:return:
		"""
		learned_action_preconditions = learned_model.actions[action_name].precondition
		learned_action_add_effects = list(learned_model.actions[action_name].effect.addlist)
		learned_action_delete_effects = list(learned_model.actions[action_name].effect.dellist)
		actual_action_preconditions = self.expected_domain.actions[action_name].precondition
		actual_action_add_effects = list(self.expected_domain.actions[action_name].effect.addlist)
		actual_action_delete_effects = list(self.expected_domain.actions[action_name].effect.dellist)
		return {
			"learned_preconditions": ",".join([str(precondition) for precondition in learned_action_preconditions]),
			"learned_add_effects": ",".join([str(add_effect) for add_effect in learned_action_add_effects]),
			"learned_delete_effects": ",".join([str(del_effect) for del_effect in learned_action_delete_effects]),
			"ground_truth_preconditions": ",".join([str(precondition) for precondition in actual_action_preconditions]),
			"ground_truth_add_effects": ",".join([str(add_effect) for add_effect in actual_action_add_effects]),
			"ground_truth_delete_effects": ",".join([str(del_effect) for del_effect in actual_action_delete_effects])
		}

	def write_learned_actions_statistics(
			self, available_trajectories: List[Trajectory], learned_model: Domain,
			learned_actions_statistics: List[Dict[str, Any]], known_actions: List[str],
			actions_observation_histogram: Dict[str, int],
			actions_with_duplicated_objects_observations: Dict[str, int],
	) -> NoReturn:
		"""Writes the statistics about the learned actions. This adds additional information about the learned actions.

		:param available_trajectories: the trajectories used in the learning process.
		:param learned_model: the action model that was learned.
		:param learned_actions_statistics: the dictionary containing the data about the learned actions.
		:param known_actions: the actions that the agent already knows.
		:param actions_with_duplicated_objects_observations: a mapping between the action and the number of times it was
			observed with duplicated objects.
		"""
		num_trajectories = len(available_trajectories)
		num_trajectory_triplets = sum([len(trajectory) for trajectory in available_trajectories])
		learned_actions = [
			action for action in learned_model.actions if action not in known_actions]
		for action_name in learned_actions:
			if action_name not in self.expected_domain.actions:
				self.logger.debug(f"Action - {action_name} is probably a proxy action. Real action not learned.")
				continue

			action_statistics = {
				"domain_name": self.expected_domain.name,
				"num_trajectories": num_trajectories,
				"num_trajectory_triplets": num_trajectory_triplets,
				"learned_action_name": action_name,
				"num_triplets_action_appeared": actions_observation_histogram[action_name],
				"num_triplets_with_duplicate_parameters": actions_with_duplicated_objects_observations[action_name]
			}
			action_statistics.update(
				self.extract_learned_action_data_vs_ground_truth(learned_model=learned_model, action_name=action_name))
			action_statistics.update(self.calculate_action_precision_recall(
				learned_model=learned_model, action_name=action_name))
			learned_actions_statistics.append(action_statistics)

		# writing the statistics for the actions that were not yet learned.
		not_yet_learned_action_names = [action_name for action_name in self.expected_domain.actions if
										action_name not in learned_model.actions]
		for not_learned_action in not_yet_learned_action_names:
			learned_actions_statistics.append(
				self.calculate_precision_recall_for_unknown_actions(
					not_learned_action, num_trajectories, num_trajectory_triplets))



"""Generates proxy actions for the ESAM algorithm."""
from itertools import combinations, product
from typing import List, Dict, Set, Any, Tuple

from pddl.pddl import Action, Effect

from sam_learner.sam_models import ComparablePredicate, SignatureType


def n_choose_k(array: List[Any], k: int, effect_type: str) -> List[Tuple[Any, ...]]:
    """Implementation of the combinatorial choosing method.

    :param array: the array containing the items that we are choosing from the subset.
    :param k: the number of elements to choose from the list.
    :param effect_type: the type of the effect, i.e. add or delete effect.
    :return: the combination of items based on the input number of elements.
    """
    fluent_combinations = list(combinations(array, k))
    return [(*combination, effect_type) for combination in fluent_combinations]


class LightProxyActionGenerator:
    """Class that is able to know whether an action contains duplicates and can create a proxy action to represent
    the inconsistent action usage."""

    def get_unit_clause_effects(
            self, effect_cnf_clauses: Dict[str, Set[ComparablePredicate]]) -> Set[ComparablePredicate]:
        """

        :param effect_cnf_clauses:
        :return:
        """
        unit_clause_effects = set()
        for cnf_effects in effect_cnf_clauses.values():
            if len(cnf_effects) == 1:
                unit_clause_effects.update(cnf_effects)

        return unit_clause_effects

    def get_precondition_fluents(
            self, precondition_cnfs: Dict[str, Set[ComparablePredicate]]) -> Set[ComparablePredicate]:
        """

        :param precondition_cnfs:
        :return:
        """
        precondition_predicates = set()
        for cnf in precondition_cnfs.values():
            precondition_predicates.update(cnf)

        return precondition_predicates

    def create_proxy_actions(self, action_name: str, action_signature: SignatureType,
                             surely_preconditions: Set[ComparablePredicate],
                             add_effect_cnfs: Dict[str, Set[ComparablePredicate]],
                             delete_effect_cnfs: Dict[str, Set[ComparablePredicate]]) -> List[Action]:
        """Creates the proxy action permutations based on the algorithm to create the power set of safe actions.

        :param action_name: the name of the original action that exists in the original domain.
        :param action_signature: the original signature of the action as known in the domain.
        :param surely_preconditions: the stored preconditions for the designated action.
        :param add_effect_cnfs: the add effects CNFs learned through the learning process.
        :param delete_effect_cnfs: the delete effects CNFs learned through the learning process.
        :return:
        """
        surely_add_effects = self.get_unit_clause_effects(add_effect_cnfs)
        surely_delete_effects = self.get_unit_clause_effects(delete_effect_cnfs)

        combined_combinations = self.create_effects_combinations(add_effect_cnfs, delete_effect_cnfs)
        effects_preconditions_product = self.create_effects_product(combined_combinations)

        proxy_actions = []
        for index, product_item in enumerate(effects_preconditions_product):
            effect_items = product_item[:-1]
            preconditions = product_item[-1]
            add_effects = []
            delete_effects = []
            for effect, effect_type in effect_items:
                if effect_type == "add-effect":
                    add_effects.append(effect)

                elif effect_type == "delete-effect":
                    delete_effects.append(effect)

            all_preconditions = list(surely_preconditions.union(preconditions))
            effect = Effect()
            effect.addlist = surely_add_effects.union(add_effects)
            effect.dellist = surely_delete_effects.union(delete_effects)
            new_action = Action(name=f"proxy-{action_name}-{index}",
                                signature=action_signature,
                                precondition=all_preconditions,
                                effect=effect)
            proxy_actions.append(new_action)

        return proxy_actions

    def create_effects_product(self, combined_combinations: List[List[Tuple[Any, ...]]]) -> List[Tuple[Any, ...]]:
        """

        :param combined_combinations:
        :return:
        """
        preconditions_effects_product = []
        selected_effects_product = list(product(*combined_combinations))
        for product_item in selected_effects_product:
            filtered_in_precondition = []
            for index, selected_effect_fluent in enumerate(product_item):
                relevant_fluents_variants = combined_combinations[index]
                fluent_preconditions_items = filter(lambda f: f != selected_effect_fluent, relevant_fluents_variants)
                fluent_preconditions = [item[0] for item in fluent_preconditions_items]
                filtered_in_precondition.extend(fluent_preconditions)

            preconditions_effects_product.append((*product_item, filtered_in_precondition))

        return preconditions_effects_product

    def create_effects_combinations(self, add_effect_cnfs, delete_effect_cnfs) -> List[List[Tuple[Any, ...]]]:
        non_unit_add_effect_fluents = [fluent_name for fluent_name in add_effect_cnfs if
                                       len(add_effect_cnfs[fluent_name]) > 1]
        non_unit_del_effect_fluents = [fluent_name for fluent_name in delete_effect_cnfs if
                                       len(delete_effect_cnfs[fluent_name]) > 1]
        combined_combinations = []
        for fluent_name in non_unit_add_effect_fluents:
            # For now we only support the case where don't remove parameters from the signature.
            combined_combinations.append(n_choose_k(list(add_effect_cnfs[fluent_name]), 1, "add-effect"))
        for fluent_name in non_unit_del_effect_fluents:
            combined_combinations.append(n_choose_k(list(delete_effect_cnfs[fluent_name]), 1, "delete-effect"))
        return combined_combinations



"""Module to handle the storing loading and updating of a trajectory."""
import logging
import os
import pickle
from pathlib import Path
from typing import Tuple, List, NoReturn, Optional

from pddl.pddl import Domain
from sam_learner.core import TrajectoryGenerator
from sam_learner.sam_models import Trajectory


class TrajectorySerializationManager:
    """class that manages the serialization processes of the trajectories."""

    working_directory_path: Path
    logger: logging.Logger
    domain_path: Path

    def __init__(self, workdir_path: Path, domain_path: Path):
        self.working_directory_path = workdir_path
        self.domain_path = domain_path
        self.logger = logging.getLogger(__name__)

    def get_problem_and_solution_files(self) -> List[Tuple[Path, Path]]:
        """Get the problem and the solution file paths from the working directory.

        :return: the paths to the problems and their respected plans.
        """
        paths = []
        for solution_file_path in self.working_directory_path.glob("*.solution"):
            problem_file_name = solution_file_path.stem.split("_plan")[0]
            problem_path = self.working_directory_path / f"{problem_file_name}.pddl"
            paths.append((problem_path, solution_file_path))

        return paths

    def create_trajectories(self, serialization_file_name: Optional[str] = None) -> List[Trajectory]:
        """Create the trajectories that will be used in the main algorithm.

        :return: the list of trajectories that will be used in the SAM algorithm.
        """
        self.logger.info("Creating the trajectories for the algorithm.")
        trajectories = []
        if serialization_file_name is not None:
            stored_trajectories_path = self.working_directory_path / serialization_file_name

        else:
            stored_trajectories_path = self.working_directory_path / "saved_trajectories"

        if stored_trajectories_path.exists():
            return self.load_trajectories(stored_trajectories_path)

        for problem_path, plan_path in self.get_problem_and_solution_files():
            generator = TrajectoryGenerator(str(self.domain_path), str(problem_path))
            trajectories.append(generator.generate_trajectory(str(plan_path)))

        self.store_trajectories(stored_trajectories_path, trajectories)
        return trajectories

    def create_trajectories_fama_format(self) -> Tuple[List[Path], List[Trajectory]]:
        """Create the trajectories in the files that FAMA learner can use in the learning process.

        :return: the list of paths to the trajectories.
        """
        self.logger.info("Creating the trajectories for the FAMA algorithm in the correct format.")
        trajectory_paths = []
        generated_trajectories = []
        for index, (problem_path, plan_path) in enumerate(self.get_problem_and_solution_files()):
            generator = TrajectoryGenerator(str(self.domain_path), str(problem_path))
            fama_trajectory, generated_trajectory = generator.create_trajectory_in_fama_format(plan_path)
            trajectory_file_path = self.working_directory_path / f"{self.domain_path.stem}_trajectory{index}"
            trajectory_paths.append(trajectory_file_path)
            generated_trajectories.append(generated_trajectory)
            with open(trajectory_file_path, 'w') as output:
                output.write(fama_trajectory)

        return trajectory_paths, generated_trajectories

    def load_trajectories(self, stored_trajectories_path: Path) -> List[Trajectory]:
        """Loads the trajectories from the trajectories file.

        :param stored_trajectories_path: the path to the files that stores the trajectories.
        :return: the loaded deserialized trajectories.
        """
        self.logger.debug("Loading the trajectories from the file!")
        with open(stored_trajectories_path, "rb") as trajectories_file:
            return pickle.load(trajectories_file)

    def store_trajectories(self, stored_trajectories_path: Path, trajectories: List[Trajectory]) -> NoReturn:
        """Store the trajectories in the trajectory file so that future runs of the algorithm would be faster.

        :param stored_trajectories_path: the path to the file that stores the trajectories.
        :param trajectories: the trajectories that are to be stored in the file.
        """
        with open(stored_trajectories_path, "wb") as trajectories_file:
            self.logger.debug("Saving the created trajectories in a file for future usage.")
            pickle.dump(trajectories, trajectories_file)

    def update_stored_trajectories(self, trajectory: Trajectory, save_path: Optional[Path] = None) -> NoReturn:
        """Serialize the new trajectory.

        :param trajectory: the trajectory to serialize.
        :param save_path: the path to the file that saves the trajectories.
        """
        trajectories_path = self.working_directory_path / "saved_trajectories" if save_path is None else save_path
        with open(trajectories_path, "rb") as trajectories_file:
            trajectories = pickle.load(trajectories_file)

        trajectories.append(trajectory)
        self.store_trajectories(trajectories_path, trajectories)

    def delete_trajectories_file(self, trajectories_file_path: Optional[Path] = None) -> NoReturn:
        """deletes the file that contains the saved trajectories.

        :param trajectories_file_path: the path to the trajectories file.
        """
        trajectories_path = self.working_directory_path / "saved_trajectories" if \
            trajectories_file_path is None else trajectories_file_path
        os.remove(trajectories_path)



"""Module that converts a problem data and a plan to a trajectory."""
import logging
from typing import List, Tuple, NoReturn

import grounding
from pathlib import Path
from pddl.pddl import Problem, Domain, Action, Predicate
from pyperplan import Parser
from task import Operator

from sam_learner.core.grounded_action_locator import parse_plan_action_string, locate_lifted_action, \
    ground_lifted_action
from sam_learner.sam_models.parameter_binding import ParameterBinding
from sam_learner.sam_models.state import State
from sam_learner.sam_models.trajectory_component import TrajectoryComponent
from sam_learner.sam_models.types import Trajectory


class TrajectoryGenerator:
    """Class that generates trajectory out of a problem file and it's corresponding plan."""

    logger: logging.Logger
    domain: Domain
    problem: Problem

    def __init__(self, domain_path: str, problem_path: str):
        parser = Parser(domain_path, problem_path)
        self.logger = logging.getLogger(__name__)
        self.domain = parser.parse_domain(read_from_file=True)
        self.problem = parser.parse_problem(dom=self.domain, read_from_file=True)

    def _parse_plan_actions(self, plan_path: str) -> Tuple[List[str], List[Action]]:
        """Parse the actions that exist in the plan file and extract the action objects with the
        object binding.

        :param plan_path: the path to the file containing the plan that solves the input problem.
        :return: both the grounded operator names and the actions to execute.
        """
        self.logger.debug(f"Parsing the plan file in the path - {plan_path}")
        with open(plan_path, "rt") as plan_file:
            lines = plan_file.readlines()
            lines = [line.strip("\n") for line in lines]
            if "cost" in lines[-1]:
                lines.pop()  # remove the cost line from the plan.

        actions = []
        self.logger.debug("The plan contains the following grounded actions:")
        for index, line in enumerate(lines):
            self.logger.debug(line)
            action_name, action_params = parse_plan_action_string(line)
            if len(action_params) == 0:
                lines[index] = f"({action_name})"

            actions.append(locate_lifted_action(self.domain, action_name))

        return lines, actions

    def _create_state_parameter_matching(self, grounded_state: frozenset) -> State:
        """Matches between the grounded and the lifted predicates.

        :param grounded_state: the grounded set of facts that represent the state.
        :return: a match between the grounded and the lifted predicates.
        """
        lifted_state_data: List[Tuple[Predicate, List[ParameterBinding]]] = \
            State.generate_problem_state(grounded_state, self.domain, self.problem)
        grounded_predicates = []
        for predicate, bindings in lifted_state_data:
            signature = [binding.bind_parameter() for binding in bindings]
            grounded_predicates.append(Predicate(predicate.name, signature))

        return State(grounded_predicates, self.domain)

    def _create_single_trajectory_component(
            self, index: int, action: Action, op_name: str,
            operators: dict, previous_state: State) -> TrajectoryComponent:
        """Create a single trajectory component by applying the action on the previous state.

        :param index: the index of the step that is being parsed currently.
        :param action: the lifted action that is to be executed on the state.
        :param op_name: the grounded operator's name.
        :param operators: the grounded operators that can apply an action on a state.
        :param previous_state: the previous state to be changed.
        :return: the trajectory component representing the current stage.
        """
        grounded_action: Operator = operators[op_name]
        grounded_next_state_statements = grounded_action.apply(previous_state.ground_facts())
        next_state = self._create_state_parameter_matching(grounded_next_state_statements)
        _, action_objects = parse_plan_action_string(grounded_action.name)
        _, bindings = ground_lifted_action(action, action_objects)
        component = TrajectoryComponent(index=index,
                                        previous_state=previous_state,
                                        action=action,
                                        action_parameters_binding=bindings,
                                        next_state=next_state)
        self.logger.debug(f"Finished creating the component:\n{component}")
        return component

    def generate_trajectory(self, plan_path: str, should_return_partial_trajectory: bool = False) -> Trajectory:
        """Generates a trajectory out of a problem file and a plan file.

        :param plan_path the path to the plan generated by a solver.
        :param should_return_partial_trajectory: whether or not to return a partial trajectory in case of failure.
        :return: the trajectory that represents the plan file.

        Note:
            the plan output should be in the form of: (load-truck obj23 tru2 pos2) ...
        """
        self.logger.info(f"Generating a trajectory from the file: {self.problem}")
        trajectory = []
        grounded_planning_task = grounding.ground(problem=self.problem)
        operators = {operator.name: operator for operator in grounded_planning_task.operators}
        op_names, plan_actions_sequence = self._parse_plan_actions(plan_path)
        previous_state = State(self.problem.initial_state, self.domain)

        self.logger.info("Starting to iterate over the actions sequence.")
        for index, (op_name, action) in enumerate(zip(op_names, plan_actions_sequence)):
            try:
                component = self._create_single_trajectory_component(
                    index, action, op_name, operators, previous_state)
                trajectory.append(component)
                previous_state = component.next_state

            except AssertionError:
                error_message = f"The operation {op_name} is not applicable! The failed action - {action.name}"
                self.logger.warning(error_message)
                if should_return_partial_trajectory:
                    self.logger.debug(f"Returning partial trajectory since the flag was turned on.")
                    return trajectory

                raise AssertionError(error_message)

        return trajectory

    def _format_trajectory_fama_problem_objects(self) -> str:
        """Formats the trajectory's problem objects in the format that the algorithm can read."""
        problem_objects = self.problem.objects
        return " ".join([f"{object_name} - {object_type.name}" for object_name, object_type in problem_objects.items()])

    def create_trajectory_in_fama_format(self, plan_path: Path) -> Tuple[str, Trajectory]:
        """Create the trajectory output in a format that FAMA algorithm can use to read.

        :param plan_path: the path to the plan file of the current problem.
        :return: the string representation of the trajectory in the format that FAMA can easily read.
        """
        trajectory = self.generate_trajectory(str(plan_path))
        trajectory_str = ""
        trajectory_str += f"(trajectory\n\n(:objects {self._format_trajectory_fama_problem_objects()})\n\n"
        initial_state = trajectory[0].previous_state
        trajectory_str += f"(:init {' '.join([fact for fact in initial_state.ground_facts()])})\n\n"
        for trajectory_component in trajectory:
            trajectory_str += f"(:action {trajectory_component.grounded_action.activated_action_representation})\n\n"
            trajectory_str += f"(:state " \
                              f"{' '.join([fact for fact in trajectory_component.next_state.ground_facts()])})\n\n"

        trajectory_str += ")"
        return trajectory_str, trajectory

    def validate_trajectory(self, trajectory: Trajectory) -> bool:
        """Validate that the last state in the trajectory is indeed the goal state.

        :param trajectory: the trajectory to validate.
        :return: whether or not the trajectory ends with the goal state.
        """
        last_component: TrajectoryComponent = trajectory[-1]
        last_state = last_component.next_state
        grounded_goals = grounding.ground(problem=self.problem).goals
        return grounded_goals <= last_state.ground_facts()



"""An extension to the predicate matcher that is able to cope with ambiguity in the executed actions."""
from itertools import permutations
from typing import List, Tuple, Optional

from pddl.pddl import Type, Domain

from sam_learner.sam_models import ComparablePredicate, SignatureType, GroundedAction
from .predicates_matcher import PredicatesMatcher


def create_signature_permutations(
        grounded_signature: SignatureType, lifted_signature: SignatureType,
        subset_size: int) -> List[Tuple[Tuple[str, Tuple[Type]], Tuple[str, Tuple[Type]]]]:
    """Choose r items our of a list size n.

    :param grounded_signature: the action's grounded signature.
    :param lifted_signature: the action's lifted signature.
    :param subset_size: the size of the subset.
    :return: a list containing subsets of the original list.
    """
    matching_signatures = zip(grounded_signature, lifted_signature)
    matching_permutations = list(permutations(matching_signatures, subset_size))

    return matching_permutations


class ExtendedMatcher(PredicatesMatcher):
    """An extended version of the predicate matcher class."""

    def __init__(self, domain_path: Optional[str] = None, domain: Optional[Domain] = None):
        super().__init__(domain_path, domain)

    def _is_matching_possible(self, grounded_signature: SignatureType, predicate_objects: List[str]) -> bool:
        """Test whether it is possible to match the predicate to the current action.

        :param grounded_signature: the action's grounded signature.
        :param predicate_objects: the names of the objects that appear in the predicate.
        :return: whether or not it is possible to match the predicate to the action based on the action's signature.
        """
        grounded_action_objects = [signature_item[0] for signature_item in grounded_signature]
        constants_names = [name for name in self.matcher_domain.constants]

        possible_grounded_matches = grounded_action_objects + constants_names
        if not all(predicate_obj in possible_grounded_matches for predicate_obj in predicate_objects):
            self.logger.debug("The predicate objects are not contained in the action's object, matching aborted.")
            return False

        return True

    def extended_predicate_matching(
            self, grounded_predicate: ComparablePredicate, grounded_signature: SignatureType,
            lifted_signature: SignatureType) -> Optional[List[ComparablePredicate]]:
        """The extended functionality that matches predicates to actions with duplicates.

        :param grounded_predicate: the grounded predicate that appeared in the trajectory.
        :param grounded_signature: the action's grounded signature.
        :param lifted_signature: the action's lifted signature.
        :return: the possible matching predicates.
        """
        predicate_objects = [signature_item[0] for signature_item in grounded_predicate.signature]
        if not self._is_matching_possible(grounded_signature, predicate_objects):
            return None

        constant_signature_items = [(name, (const_type,)) for name, const_type in self.matcher_domain.constants.items()]
        grounded_objects = grounded_signature + constant_signature_items
        lifted_objects = lifted_signature + constant_signature_items
        matching_signature_permutations = create_signature_permutations(
            grounded_objects, lifted_objects, len(predicate_objects))

        possible_matches = []
        for signature_option in matching_signature_permutations:
            lifted_match = []
            matching_grounded_action_objects = []
            for grounded_signature_item, lifted_signature_item in signature_option:
                lifted_match.append(lifted_signature_item)
                matching_grounded_action_objects.append(grounded_signature_item[0])

            if predicate_objects == matching_grounded_action_objects:
                possible_matches.append(ComparablePredicate(grounded_predicate.name, lifted_match))

        return possible_matches

    def match_predicate_to_action_literals(
            self, grounded_predicate: ComparablePredicate,
            grounded_signature: SignatureType,
            lifted_signature: SignatureType) -> Optional[List[ComparablePredicate]]:
        """Match a literal to a possible lifted precondition for the input action.

        Note:
            This method does not raise an error in case that there are duplications in either the state or the action.
            This method first tries to use the parent matching method. In case of an error being raised, the method
            will then rollback to the extended matching procedure.

        :param grounded_predicate: the grounded predicate that represent part of the previous state.
        :param grounded_signature: the signature of the action that contains the actual objects
            that the action was executed on.
        :param lifted_signature: the lifted signature of the action, is accessible from the
            trajectory.
        :return: a precondition, in case the action and the predicate contain matching objects,
            None otherwise.
        """
        try:
            match = super(ExtendedMatcher, self).match_predicate_to_action_literals(
                grounded_predicate, grounded_signature, lifted_signature)
            return None if match is None else [match]

        except ValueError:
            self.logger.debug("Found duplications in either the state of the action. Due to this fact, "
                              "rolling back to the extend matching procedure.")
            return self.extended_predicate_matching(grounded_predicate, grounded_signature, lifted_signature)

    def get_possible_literal_matches(
            self, grounded_action: GroundedAction, literals: List[ComparablePredicate]) -> List[ComparablePredicate]:
        """Get a list of possible preconditions for the action according to the previous state.

        :param grounded_action: the grounded action that was executed according to the trajectory.
        :param literals: the list of literals that we try to match according to the action.
        :return: a list of possible preconditions for the action that is being executed.
        """
        possible_matches = []
        lifted_signature = grounded_action.lifted_signature
        grounded_signature = grounded_action.grounded_signature
        for predicate in literals:
            matches = self.match_predicate_to_action_literals(predicate, grounded_signature, lifted_signature)
            if matches is None:
                continue

            possible_matches.extend(matches)

        return possible_matches



from distutils.core import setup
from setuptools import find_packages

with open("README.md") as f:
    long_description = f.read()

setup(name='sam-learner',
      version='2.1.2',
      python_requires=">=3.8",
      description='Safe Action Model Learner',
      long_description=long_description,
      long_description_content_type="text/markdown",
      keywords="automatic safe action model learning",
      author='Argaman Mordoch',
      packages=find_packages(exclude=["tests", "utils"]),
     )